R"(
/* ************************************************************************
 * Copyright 2015 Vratis, Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ************************************************************************ */

//do not remove. This kernel is used to measure some parameters of the device
#ifndef WG_SIZE
#error WG_SIZE undefined!
#endif

__kernel
__attribute__((reqd_work_group_size(WG_SIZE,1,1)))
void control( const int bogus )
{
    return;
}
)"
